package io.gitlab.jfronny.respackopts.util;

import io.gitlab.jfronny.respackopts.Respackopts;
import io.gitlab.jfronny.respackopts.model.DirRpo;
import io.gitlab.jfronny.respackopts.model.FileRpo;
import io.gitlab.jfronny.respackopts.model.PackMeta;
import io.gitlab.jfronny.respackopts.model.ThrowingBiConsumer;
import io.gitlab.jfronny.respackopts.model.enums.ConfigSyncMode;
import io.gitlab.jfronny.respackopts.model.enums.PackCapability;
import io.gitlab.jfronny.respackopts.model.tree.ConfigBranch;
import net.minecraft.resource.ResourcePack;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;

public class MetaCache {
    private static final Map<Path, ConfigBranch> CONFIG_BRANCH = new HashMap<>(); // pack.rpo -> active config branch
    private static final Map<Path, PackMeta> PACK_METAS = new HashMap<>(); // pack.rpo -> original pack metadata
    private static final Map<Path, String> PACK_IDS = new HashMap<>(); // pack.rpo -> pack id
    private static final Map<String, Path> DISPLAY_NAME_LOOKUP = new HashMap<>(); // display name -> pack.rpo
    private static final Map<String, Path> PACK_NAME_LOOKUP = new HashMap<>(); // pack name -> pack.rpo
    public static final Map<Path, Map<String, DirRpo>> DIR_RPOS = new HashMap<>(); // pack.rpo -> .rpo path -> parsed .rpo
    public static final Map<Path, Map<String, FileRpo>> FILE_RPOS = new HashMap<>(); // pack.rpo -> .rpo path -> parsed .rpo
    public static void remove(Path packRpo) {
        CONFIG_BRANCH.remove(packRpo);
        PACK_METAS.remove(packRpo);
        PACK_IDS.remove(packRpo);
        //DISPLAY_NAME_LOOKUP
        //PACK_NAME_LOOKUP
        DIR_RPOS.remove(packRpo);
        FILE_RPOS.remove(packRpo);
    }

    public static void addFromScan(String displayName, String packName, PackMeta meta, Path dataLocation) {
        if (Respackopts.META_VERSION > meta.version) {
            Respackopts.LOGGER.warn(displayName + " uses an outdated RPO format (" + meta.version + "). Although this is supported, using the latest version (" + Respackopts.META_VERSION + ") is recommended");
        }
        meta.conf.setVersion(meta.version);
        if (meta.version < 5) meta.capabilities.add(PackCapability.DirFilter);
        remove(dataLocation);
        if (CONFIG_BRANCH.containsKey(dataLocation))
            CONFIG_BRANCH.get(dataLocation).sync(meta.conf, ConfigSyncMode.RESPACK_LOAD);
        else
            CONFIG_BRANCH.put(dataLocation, meta.conf.clone());
        DISPLAY_NAME_LOOKUP.put(displayName, dataLocation);
        PACK_NAME_LOOKUP.put(packName, dataLocation);
        PACK_METAS.put(dataLocation, meta);
        DIR_RPOS.put(dataLocation, new HashMap<>());
        FILE_RPOS.put(dataLocation, new HashMap<>());
        PACK_IDS.put(dataLocation, meta.id);
        if (!dataLocation.startsWith(Respackopts.FALLBACK_CONF_DIR)) {
            Path legacyLocation = Respackopts.FALLBACK_CONF_DIR.resolve(meta.id + ".json");
            if (Files.exists(legacyLocation) && !Files.exists(dataLocation)) {
                try {
                    Files.move(legacyLocation, dataLocation);
                } catch (IOException e) {
                    Respackopts.LOGGER.error("Could not move data to new location", e);
                }
            }
        }
        load(dataLocation);
        save(dataLocation, meta.conf);
    }

    public static void save() {
        if (Respackopts.CONFIG.debugLogs)
            Respackopts.LOGGER.info("Saving configs");
        for (Map.Entry<Path, ConfigBranch> e : CONFIG_BRANCH.entrySet()) {
            save(e.getKey(), e.getValue());
        }
        for (Runnable action : Respackopts.SAVE_ACTIONS) {
            action.run();
        }
    }

    private static void save(Path dataLocation, ConfigBranch branch) {
        try (Writer writer = Files.newBufferedWriter(dataLocation)) {
            Respackopts.GSON.toJson(branch, writer);
            writer.flush();
        } catch (IOException ex) {
            Respackopts.LOGGER.error("Could not save config", ex);
        }
    }

    public static void load(Path dataLocation) {
        if (Files.exists(dataLocation)) {
            if (Respackopts.CONFIG.debugLogs)
                Respackopts.LOGGER.info("Loading configs for: " + dataLocation);
            try (Reader reader = Files.newBufferedReader(dataLocation)) {
                ConfigBranch b = Respackopts.GSON.fromJson(reader, ConfigBranch.class);
                if (CONFIG_BRANCH.containsKey(dataLocation))
                    CONFIG_BRANCH.get(dataLocation).sync(b, ConfigSyncMode.CONF_LOAD);
                Respackopts.STAR_SCRIPT.set(Respackopts.sanitizeString(getId(dataLocation)), b::buildStarScript);
            } catch (IOException e) {
                Respackopts.LOGGER.error("Failed to load " + dataLocation, e);
            }
        }
    }

    public static Path getDataLocation(ResourcePack pack) {
        return PACK_NAME_LOOKUP.get(pack.getName());
    }

    public static PackMeta getMeta(Path dataLocation) {
        return PACK_METAS.get(dataLocation);
    }

    public static @Nullable Path getDataLocationByDisplayName(String displayName) {
        if (!DISPLAY_NAME_LOOKUP.containsKey(displayName)) return null;
        return DISPLAY_NAME_LOOKUP.get(displayName);
    }

    public static ConfigBranch getBranch(Path dataLocation) {
        return CONFIG_BRANCH.get(dataLocation);
    }

    public static String getId(Path dataLocation) {
        return PACK_IDS.get(dataLocation);
    }

    public static boolean hasCapability(ResourcePack pack, PackCapability capability) {
        Path dataLocation = getDataLocation(pack);
        if (dataLocation == null) return false;
        if (!PACK_METAS.containsKey(dataLocation)) {
            StringBuilder sb = new StringBuilder("Could not get pack with \"");
            sb.append(dataLocation);
            sb.append("\" (available: ");
            for (Path path : PACK_METAS.keySet()) {
                sb.append(path).append(", ");
            }
            throw new NullPointerException(sb.substring(0, sb.length() - 2) + ')');
        }
        return PACK_METAS.get(dataLocation).capabilities.contains(capability);
    }

    public static <TEx extends Exception> void forEach(ThrowingBiConsumer<Path, ConfigBranch, TEx> idAndBranchConsumer) throws TEx {
        for (Map.Entry<Path, ConfigBranch> entry : CONFIG_BRANCH.entrySet()) {
            idAndBranchConsumer.accept(entry.getKey(), entry.getValue());
        }
    }
}
